# isort: skip_file

from dotenv import load_dotenv

__version__ = "0.0.2"
load_dotenv()

# Ignore import sorting, since LLMEngine needs to be imported first
from grazier.engines.llm import LLMEngine  # noqa: F401, E402
from grazier.engines.chat import Conversation, LLMChat, Speaker  # noqa: F401, E402



from typing import Optional

import torch


def select_device(device: Optional[str] = None) -> Optional[str]:
    if device is None:
        # Check if CUDA is available
        if torch.cuda.is_available():
            device = "cuda"
        else:
            device = "cpu"

    if isinstance(device, str):
        if device == "cuda":
            # If the user specifies only CUDA, use the GPU with the most free memory
            free_memory = [
                torch.cuda.get_device_properties(i).total_memory - torch.cuda.memory_allocated(i)
                for i in range(torch.cuda.device_count())
            ]
            device = "cuda:" + str(torch.argmax(torch.tensor(free_memory)).item())
        elif device.startswith("cuda:"):
            # If the user specifies a specific GPU, make sure it exists
            if int(device[5:]) >= torch.cuda.device_count():
                raise ValueError(f"Invalid device: {device}")
        elif device == "api" or device == "defer":
            return None

        return device

    elif isinstance(device, int):
        return f"cuda:{device}"

    return None



try:
    from vertexai.preview.language_models import ChatModel, InputOutputTextPair
except ImportError:
    ChatModel = None
    InputOutputTextPair = None

from typing import Any, Dict, List

from ratelimit import limits, sleep_and_retry

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import singleton


class VertexLLMEngine(LLMChat):
    def __init__(self, model: str, **kwargs: Dict[str, Any]) -> None:
        super().__init__(device="api")

        if ChatModel is None:
            raise ImportError("Please install the Vertex AI SDK to use this LM engine.")

        self._model = ChatModel.from_pretrained(model)
        self._parameters = {
            "max_output_tokens": 256,  # Token limit determines the maximum amount of text output.
            "top_p": 0.95,  # Tokens are selected from most probable to least until the sum of their probabilities equals the top_p value.
            "top_k": 40,  # A top_k of 1 means the selected token is the most probable among all tokens.
        } | kwargs

    @sleep_and_retry  # type: ignore
    @limits(  # type: ignore
        calls=40, period=60
    )  # This is the default rate limit for Vertex AI (actual rate limit is 60 calls per minute, but we'll be conservative)
    def _rate_limited_model_predict(self, info, **kwargs: Any) -> Any:
        context, examples, prompt = info
        chat = self._model.start_chat(context=context, examples=examples, **kwargs)
        response = chat.send_message(prompt).text

        return response

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        # Start the chat
        system_turns = [c for c in conversation.turns if c.speaker == Speaker.SYSTEM]
        context = system_turns[-1].text if system_turns else ""
        non_system_turns = [c for c in conversation.turns if c.speaker != Speaker.SYSTEM]

        # Assert that the non-system turns alternate between the user and the agent
        for idx, turn in enumerate(non_system_turns):
            if idx % 2 == 0:
                assert turn.speaker == Speaker.USER
            else:
                assert turn.speaker == Speaker.AI
        if len(non_system_turns) > 1:
            assert non_system_turns[-1].speaker == Speaker.USER
            assert InputOutputTextPair is not None
            # Build the examples
            examples = [
                InputOutputTextPair(input_text=non_system_turns[i].text, output_text=non_system_turns[i + 1].text)
                for i in range(0, len(non_system_turns) - 1, 2)
            ]
        else:
            examples = []

        # Normalize kwargs from openai to vertexai (some common parameters are different)
        kwargs = (
            self._parameters
            | {
                "max_output_tokens": kwargs.pop("max_output_tokens", kwargs.pop("max_tokens", 256)),
                "temperature": kwargs.pop("temperature", 1.0),
            }
            | kwargs
        )

        return [
            ConversationTurn(
                self._rate_limited_model_predict((context, examples, non_system_turns[-1].text), **kwargs),
                speaker=Speaker.AI,
            )  # type: ignore
            for _ in range(n_completions)
        ]


@register_engine
@singleton
class PaLMEngine(VertexLLMEngine):
    name = ("PaLM", "palm")

    def __init__(self) -> None:
        super().__init__("chat-bison@001")



try:
    from vertexai.preview.language_models import TextGenerationModel
except ImportError:
    TextGenerationModel = None

from typing import Any, Dict, List

from ratelimit import limits, sleep_and_retry

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.python import singleton


class VertexLLMEngine(LLMEngine):
    def __init__(self, model: str, **kwargs: Dict[str, Any]) -> None:
        if TextGenerationModel is None:
            raise ImportError("Please install the Vertex AI SDK to use this LM engine.")

        self._model = TextGenerationModel.from_pretrained(model)
        self._parameters = {
            # Token limit determines the maximum amount of text output.
            "max_output_tokens": kwargs.pop("max_output_tokens", kwargs.pop("max_tokens", 256)),
            "top_p": 0.95,  # Tokens are selected from most probable to least until the sum of their probabilities equals the top_p value.
            "top_k": 40,  # A top_k of 1 means the selected token is the most probable among all tokens.
        } | kwargs

    @sleep_and_retry  # type: ignore
    @limits(  # type: ignore
        calls=40, period=60
    )  # This is the default rate limit for Vertex AI (actual rate limit is 60 calls per minute, but we'll be conservative)
    def _rate_limited_model_predict(self, *args: Any, **kwargs: Any) -> Any:
        return self._model.predict(*args, **kwargs)

    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        # Normalize kwargs from openai to vertexai (some common parameters are different)
        kwargs = (
            self._parameters
            | {
                "max_output_tokens": kwargs.pop("max_output_tokens", kwargs.pop("max_tokens", 256)),
                "temperature": kwargs.pop("temperature", 1.0),
            }
            | kwargs
        )

        return [self._rate_limited_model_predict(prompt, **kwargs).text for _ in range(n_completions)]  # type: ignore


@register_engine
@singleton
class PaLMEngine(VertexLLMEngine):
    name = ("PaLM", "palm")

    def __init__(self) -> None:
        super().__init__("text-bison@001")



import logging
import random

import pytest

from grazier.engines.llm import LLMEngine


@pytest.mark.parametrize("engine", ["j2-light", "j2-mid", "j2-ultra"])
def test_ai21_llm_engine(engine: str) -> None:
    _engine = LLMEngine.from_string(engine)
    random_number = random.randint(0, 100)
    response = _engine(f"My name, followed by a colon with the number {random_number} is:")

    for r in response:
        if len(r.strip()) == 0:
            pytest.skip(f"Empty response from {engine} engine")
        if str(random_number) not in r:
            logging.warning(f'Number "{random_number}" not found in response "{r}"')



import click


@click.group()
def main() -> None:
    pass


@main.command()
def list_engines() -> None:
    """List available engines."""
    from grazier.engines.chat import LM_CHAT_ENGINES_CLI
    from grazier.engines.llm import LM_ENGINES_CLI

    print("Available LLM engines (Completion):")

    # Types here are super broken because of the wrapped decorator
    for k, v in sorted(LM_ENGINES_CLI.items()):  # type: ignore
        print(f"\t - {v.name[0]}/{k}")  # type: ignore

    print("Available LLM engines (Chat):")
    # Types here are super broken because of the wrapped decorator
    # Types here are super broken because of the wrapped decorator
    for k, v in sorted(LM_CHAT_ENGINES_CLI.items()):  # type: ignore
        print(f"\t - {v.name[0]}/{k}")  # type: ignore


@main.command()
@click.argument("engine")
@click.argument("prompt")
def complete(
    engine: str,
    prompt: str,
) -> None:
    from grazier import LLMEngine

    _e = LLMEngine.from_string(engine)
    print(_e(prompt, n_completions=2)[0])



from typing import Any, List, Type

from grazier.engines.chat import Conversation, Speaker
from grazier.engines.llm import LLMEngine


def wrap_chat_llm_engine(cls) -> Type[LLMEngine]:  # type: ignore
    """
    Wrap an LLMChat engine to make it compatible with the LLMEngine interface.

    Args:
        cls (Type[LLMChat]): The LLMChat engine to wrap.

    Returns:
        Type[LLMEngine]: The wrapped LLMEngine.
    """

    class _WrappedEngine(LLMEngine):
        name = (f"{cls.name[0]} (Chat)", f"{cls.name[1]}")

        def __init__(self, **kwargs: Any) -> None:
            super().__init__(device="defer")
            self._engine = cls(**kwargs)

        def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
            # Build a conversation
            conversation = Conversation()
            conversation.add_turn(prompt, speaker=Speaker.USER)
            return [x.text for x in self._engine.call(conversation, n_completions=n_completions, **kwargs)]

    return _WrappedEngine



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

from grazier.engines.llm import register_chat_engine
from grazier.utils.pytorch import select_device


class Speaker(Enum):
    USER = "user"
    AI = "ai"
    SYSTEM = "system"


@dataclass
class ConversationTurn:
    text: str
    speaker: Speaker


class Conversation:
    def __init__(self, turns: Optional[List[Union[str, ConversationTurn]]] = None):
        if turns is not None:
            # Convert strings to ConversationTurns
            self.turns = [ConversationTurn(text=t, speaker=Speaker.USER) if isinstance(t, str) else t for t in turns]
        else:
            self.turns = []

    def add_turn(self, text: Union[str, ConversationTurn], speaker: Optional[Speaker] = None) -> None:
        if isinstance(text, str):
            self.turns.append(ConversationTurn(text=text, speaker=speaker or Speaker.USER))
        else:
            self.turns.append(ConversationTurn(text=text.text, speaker=speaker or text.speaker))

    def __repr__(self) -> str:
        # TODO: Make this more readable
        return f"Conversation({self.turns})"

    def __len__(self) -> int:
        return len(self.turns)

    def __getitem__(self, i: int) -> ConversationTurn:
        return self.turns[i]

    def __iter__(self):
        return iter(self.turns)

    def __add__(self, other: Union[str, ConversationTurn, "Conversation"]) -> "Conversation":
        if isinstance(other, str):
            return Conversation([*self.turns, ConversationTurn(text=other, speaker=Speaker.USER)])
        elif isinstance(other, ConversationTurn):
            return Conversation([*self.turns, other])
        else:
            assert isinstance(other, Conversation)
            return Conversation(self.turns + other.turns)

    def __radd__(self, other: Union[str, ConversationTurn, "Conversation"]) -> "Conversation":
        if isinstance(other, str):
            return Conversation([ConversationTurn(text=other, speaker=Speaker.USER), *self.turns])
        elif isinstance(other, ConversationTurn):
            return Conversation([other, *self.turns])
        else:
            assert isinstance(other, Conversation)
            return Conversation(other.turns + self.turns)

    def __iadd__(self, other: Union[str, ConversationTurn, "Conversation"]) -> "Conversation":
        if isinstance(other, str):
            self.turns.append(ConversationTurn(text=other, speaker=Speaker.USER))
        elif isinstance(other, ConversationTurn):
            self.turns.append(other)
        else:
            assert isinstance(other, Conversation)
            self.turns.extend(other.turns)
        return self


class LLMChat(ABC):
    @property
    @abstractmethod
    def name(self) -> Tuple[str, str]:
        """Returns a tuple of (Pretty Name, CLI name) of the language model."""
        raise NotImplementedError()

    def __init__(self, device: Optional[str] = None) -> None:
        self.device = select_device(device)

    def __call__(
        self, prompt: Union[Conversation, ConversationTurn, str], n_completions: int = 1, **kwargs: Any
    ) -> List[ConversationTurn]:
        conversation = None
        if isinstance(prompt, str):
            conversation = Conversation()
            conversation.add_turn(prompt, speaker=Speaker.USER)
        elif isinstance(prompt, ConversationTurn):
            conversation = Conversation()
            conversation.add_turn(prompt)
        else:
            assert isinstance(prompt, Conversation)
            conversation = prompt

        return self.call(conversation, n_completions=n_completions, **kwargs)

    @abstractmethod
    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        raise NotImplementedError()

    def __repr__(
        self,
    ) -> str:
        return f"{self.__class__.__name__}({self.name[0]})"

    @classmethod
    def configure(cls, *args, **kwargs):
        raise NotImplementedError("This engine does not support automated configuration.")

    @staticmethod
    def from_string(typestr: str, **kwargs: Any) -> "LLMChat":
        typestr = typestr.lower()
        if typestr in LM_CHAT_ENGINES:
            return LM_CHAT_ENGINES[typestr](**kwargs)  # type: ignore
        elif typestr in LM_CHAT_ENGINES_CLI:
            return LM_CHAT_ENGINES_CLI[typestr](**kwargs)  # type: ignore

        raise ValueError(f"Invalid language model type: {typestr}. Valid types are: {list(LM_CHAT_ENGINES_CLI.keys())}")

    @staticmethod
    def list_models() -> List[str]:
        return list(LM_CHAT_ENGINES_CLI.keys())


LM_CHAT_ENGINES: Dict[str, Type[LLMChat]] = {}
LM_CHAT_ENGINES_CLI: Dict[str, Type[LLMChat]] = {}

T = TypeVar("T")


def register_engine(cls: T) -> T:
    from grazier.engines.llm.chat_engine import wrap_chat_llm_engine

    LM_CHAT_ENGINES[cls.name[0].lower()] = cls  # type: ignore
    LM_CHAT_ENGINES_CLI[cls.name[1].lower()] = cls  # type: ignore

    # Register the engine as an LLM Engine as well
    register_chat_engine(wrap_chat_llm_engine(cls))

    return cls


from grazier.engines.chat.anthropic_engine import *  # noqa: F403, E402
from grazier.engines.chat.bard_engine import *  # noqa: F403, E402
from grazier.engines.chat.dolly import *  # noqa: F403, E402
from grazier.engines.chat.llama_engine import *  # noqa: F403, E402
from grazier.engines.chat.openai_engine import *  # noqa: F403, E402
from grazier.engines.chat.stable_lm_engine import *  # noqa: F403, E402
from grazier.engines.chat.vertex_engine import *  # noqa: F403, E402



import os
from abc import abstractmethod
from typing import Any, List

import openai

from grazier.engines.llm import LLMEngine, register_engine
from grazier.utils.python import retry, singleton

# Setup openai api keys
openai.organization = os.getenv("OPENAI_API_ORG", None)
openai.api_key = os.getenv("OPENAI_API_KEY", None)


@singleton
class OpenAI:
    USAGE = 0.0


class OpenAICompletionLLMEngine(LLMEngine):
    @property
    @abstractmethod
    def cost_per_token(self) -> float:
        raise NotImplementedError()

    def __init__(self, model: str):
        self._model = model
        super().__init__(device="api")

    @retry(no_retry_on=(openai.error.AuthenticationError,))
    def call(self, prompt: str, n_completions: int = 1, **kwargs: Any) -> List[str]:
        cp = openai.Completion.create(model=self._model, prompt=prompt, n=n_completions, **kwargs)  # type: ignore
        OpenAI.USAGE += int(cp.usage.total_tokens) * self.cost_per_token
        return [i.text for i in cp.choices]  # type: ignore


@register_engine
@singleton
class GPT3Davinci3(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Davinci", "gpt3-davinci3")

    def __init__(self) -> None:
        super().__init__("text-davinci-003")


@register_engine
@singleton
class GPT3Davinci2(OpenAICompletionLLMEngine):
    cost_per_token = 0.03 / 1000
    name = ("GPT-3 Davinci", "gpt3-davinci2")

    def __init__(self) -> None:
        super().__init__("text-davinci-002")


@register_engine
@singleton
class GPT3Curie(OpenAICompletionLLMEngine):
    cost_per_token = 0.002 / 1000
    name = ("GPT-3 Curie", "gpt3-curie")

    def __init__(self) -> None:
        super().__init__("text-curie-001")


@register_engine
@singleton
class GPT3Babbage(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Babbage", "gpt3-babbage")

    def __init__(self) -> None:
        super().__init__("text-babbage-001")


@register_engine
@singleton
class GPT3Ada(OpenAICompletionLLMEngine):
    cost_per_token = 0.02 / 1000
    name = ("GPT-3 Ada", "gpt3-ada")

    def __init__(self) -> None:
        super().__init__("text-ada-001")



import logging
from typing import Any, List, Optional

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, StoppingCriteria, StoppingCriteriaList

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import singleton
from grazier.utils.pytorch import select_device


class StopOnTokens(StoppingCriteria):
    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs: Any) -> bool:
        stop_ids = [50278, 50279, 50277, 1, 0]
        for stop_id in stop_ids:
            if input_ids[0][-1] == stop_id:
                return True
        return False


class StableLMChatEngine(LLMChat):
    SYSTEM_PROMPT = """<|SYSTEM|># StableLM Tuned (Alpha version)
- StableLM is a helpful and harmless open-source AI language model developed by StabilityAI.
- StableLM is excited to be able to help the user, but will refuse to do anything that could be considered harmful to the user.
- StableLM is more than just an information source, StableLM is also able to write poetry, short stories, and make jokes.
- StableLM will refuse to participate in anything that could harm a human."""

    def __init__(self, model: str, device: Optional[str] = None):
        device = select_device(device)
        self._tokenizer = AutoTokenizer.from_pretrained(model)
        self._model = AutoModelForCausalLM.from_pretrained(model)
        self._model.half().to(device)
        self._device = device

    def _decode_tokens(self, tokens: torch.Tensor, last_turn_is_user: bool) -> str:
        output = self._tokenizer.decode(tokens).strip()
        if last_turn_is_user:
            output, _, _ = output.partition("<|ASSISTANT|>")
        else:
            output, _, _ = output.partition("<|USER|>")

        output = (
            output.replace("<|USER|>", "")
            .replace("<|ASSISTANT|>", "")
            .replace("<|SYSTEM|>", "")
            .replace("<|endoftext|>", "")
            .strip()
        )
        return output

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        # Build the prompt
        prompt = ""
        for idx, turn in enumerate(conversation.turns):
            if idx == 0:
                if turn.speaker != Speaker.SYSTEM:
                    # Use the default prompt if the first turn is not from the system
                    prompt += StableLMChatEngine.SYSTEM_PROMPT
                else:
                    prompt += turn.text
            elif turn.speaker == Speaker.USER:
                prompt += f"<|USER|>{turn.text}"
            elif turn.speaker == Speaker.AI:
                prompt += f"<|ASSISTANT|>{turn.text}"
            elif turn.speaker == Speaker.SYSTEM:
                logging.warning(
                    f"System turn detected at index {idx} in conversation {conversation}. This turn will be ignored."
                )

        # Add the beginning of the last turn
        last_turn_is_user = True
        if conversation.turns[-1].speaker == Speaker.AI:
            prompt += "<|USER|>"
            last_turn_is_user = False
        else:
            prompt += "<|ASSISTANT|>"
            last_turn_is_user = True

        # Actually build the completions
        inputs = self._tokenizer(prompt, return_tensors="pt").to(self._device)
        temperature = kwargs.get("temperature", None)

        if temperature is not None and temperature > 0:
            tokens = self._model.generate(
                **inputs,
                max_new_tokens=256,
                min_new_tokens=5,
                temperature=temperature,
                num_return_sequences=n_completions,
                do_sample=True,
                stopping_criteria=StoppingCriteriaList([StopOnTokens()]),
            )
        else:
            tokens = self._model.generate(
                **inputs,
                max_new_tokens=256,
                min_new_tokens=5,
                num_return_sequences=n_completions,
                do_sample=False,
                stopping_criteria=StoppingCriteriaList([StopOnTokens()]),
            )

        outputs = [self._decode_tokens(t[inputs["input_ids"].shape[-1] :], last_turn_is_user) for t in tokens]  # type: ignore
        return [
            ConversationTurn(text=output, speaker=Speaker.AI if last_turn_is_user else Speaker.USER)
            for output in outputs
        ]


@register_engine
@singleton
class StableLM3B(StableLMChatEngine):
    name = ("Stable LM (3B)", "stablelm-3b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("stabilityai/stablelm-tuned-alpha-3b", device=device)


@register_engine
@singleton
class StableLM7B(StableLMChatEngine):
    name = ("Stable LM (7B)", "stablelm-7b")

    def __init__(self, device: Optional[str] = None) -> None:
        super().__init__("stabilityai/stablelm-tuned-alpha-7b", device=device)



"""
Some of the code below is based on https://github.com/acheong08/Bard
Copyright (c) 2023 Antonio Cheong

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import asyncio
import json
import logging
import os
import random
import re
import string
from typing import Any, Dict, List, Optional

import httpx

from grazier.engines.chat import Conversation, ConversationTurn, LLMChat, Speaker, register_engine
from grazier.utils.python import retry, singleton


class Chatbot:
    """
    Synchronous wrapper for the AsyncChatbot class.
    """

    def __init__(
        self,
        session_id: str,
        proxy: Optional[dict] = None,
        timeout: int = 20,
    ):
        self.loop = asyncio.get_event_loop()
        self.async_chatbot = self.loop.run_until_complete(
            AsyncChatbot.create(session_id, proxy, timeout),
        )

    def save_conversation(self, file_path: str, conversation_name: str) -> Any:
        return self.loop.run_until_complete(
            self.async_chatbot.save_conversation(file_path, conversation_name),
        )

    def load_conversations(self, file_path: str) -> List[Dict]:
        return self.loop.run_until_complete(
            self.async_chatbot.load_conversations(file_path),
        )

    def load_conversation(self, file_path: str, conversation_name: str) -> bool:
        return self.loop.run_until_complete(
            self.async_chatbot.load_conversation(file_path, conversation_name),
        )

    @retry()
    def ask(self, message: str) -> dict:
        return self.loop.run_until_complete(self.async_chatbot.ask(message))


class AsyncChatbot:
    """
    A class to interact with Google Bard.
    Parameters
        session_id: str
            The __Secure-1PSID cookie.
        proxy: str
        timeout: int
            Request timeout in seconds.
    """

    __slots__ = [
        "headers",
        "_reqid",
        "SNlM0e",
        "conversation_id",
        "response_id",
        "choice_id",
        "proxy",
        "session_id",
        "session",
        "timeout",
    ]

    def __init__(
        self,
        session_id: str,
        proxy: Optional[dict] = None,
        timeout: int = 20,
    ):
        headers = {
            "Host": "bard.google.com",
            "X-Same-Domain": "1",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
            "Origin": "https://bard.google.com",
            "Referer": "https://bard.google.com/",
        }
        self._reqid = int("".join(random.choices(string.digits, k=4)))
        self.proxy = proxy
        self.conversation_id = ""
        self.response_id = ""
        self.choice_id = ""
        self.session_id = session_id
        self.session = httpx.AsyncClient(proxies=self.proxy)
        self.session.headers = headers  # type: ignore
        self.session.cookies.set("__Secure-1PSID", session_id)
        self.timeout = timeout

    @classmethod
    async def create(
        cls,
        session_id: str,
        proxy: Optional[dict] = None,
        timeout: int = 20,
    ) -> "AsyncChatbot":
        instance = cls(session_id, proxy, timeout)
        instance.SNlM0e = await instance.__get_snlm0e()
        return instance

    async def save_conversation(self, file_path: str, conversation_name: str) -> None:
        """
        Saves conversation to the file
        :param file_path: file to save (json)
        :param conversation_name: any name of current conversation (unique one)
        :return: None
        """
        # Load conversations from file
        conversations = await self.load_conversations(file_path)

        # Update existing one
        conversation_exists = False
        for conversation in conversations:
            if conversation["conversation_name"] == conversation_name:
                conversation["conversation_name"] = conversation_name
                conversation["_reqid"] = self._reqid
                conversation["conversation_id"] = self.conversation_id
                conversation["response_id"] = self.response_id
                conversation["choice_id"] = self.choice_id
                conversation["SNlM0e"] = self.SNlM0e
                conversation_exists = True

        # Create conversation object
        if not conversation_exists:
            conversation = {
                "conversation_name": conversation_name,
                "_reqid": self._reqid,
                "conversation_id": self.conversation_id,
                "response_id": self.response_id,
                "choice_id": self.choice_id,
                "SNlM0e": self.SNlM0e,
            }
            conversations.append(conversation)

        # Save to the file
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(conversations, f, indent=4)

    async def load_conversations(self, file_path: str) -> List[Dict]:
        # Check if file exists
        if not os.path.isfile(file_path):
            return []
        with open(file_path, encoding="utf-8") as f:
            return json.load(f)

    async def load_conversation(self, file_path: str, conversation_name: str) -> bool:
        """
        Loads a conversation from history file. Returns whether the conversation was found
        :param file_path: File with conversations (json)
        :param conversation_name: unique conversation name
        :return: True if the conversation was found
        """
        conversations = await self.load_conversations(file_path)
        for conversation in conversations:
            if conversation["conversation_name"] == conversation_name:
                self._reqid = conversation["_reqid"]
                self.conversation_id = conversation["conversation_id"]
                self.response_id = conversation["response_id"]
                self.choice_id = conversation["choice_id"]
                self.SNlM0e = conversation["SNlM0e"]
                return True
        return False

    async def __get_snlm0e(self) -> str:
        # Find "SNlM0e":"<ID>"
        if not self.session_id or self.session_id[-1] != ".":
            raise Exception(
                "__Secure-1PSID value must end with a single dot. Enter correct __Secure-1PSID value.",
            )
        resp = await self.session.get(
            "https://bard.google.com/",
            timeout=10,
            follow_redirects=True,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Response code not 200. Response Status is {resp.status_code}",
            )
        SNlM0e = re.search(r"SNlM0e\":\"(.*?)\"", resp.text)
        if not SNlM0e:
            raise Exception(
                "SNlM0e value not found in response. Check __Secure-1PSID value.",
            )
        return SNlM0e.group(1)

    async def ask(self, message: str) -> dict:
        """
        Send a message to Google Bard and return the response.
        :param message: The message to send to Google Bard.
        :return: A dict containing the response from Google Bard.
        """
        # url params
        params = {
            "bl": "boq_assistant-bard-web-server_20230620.14_p0",
            "_reqid": str(self._reqid),
            "rt": "c",
        }

        # message arr -> data["f.req"]. Message is double json stringified
        message_struct = [
            [message],
            None,
            [self.conversation_id, self.response_id, self.choice_id],
        ]
        data = {
            "f.req": json.dumps([None, json.dumps(message_struct)]),
            "at": self.SNlM0e,
        }
        resp = await self.session.post(
            "https://bard.google.com/_/BardChatUi/data/assistant.lamda.BardFrontendService/StreamGenerate",
            params=params,
            data=data,
            timeout=self.timeout,
        )
        chat_data = json.loads(resp.content.splitlines()[3])[0][2]
        if not chat_data:
            return {"content": f"Google Bard encountered an error: {resp.content!r}."}
        json_chat_data = json.loads(chat_data)
        images = []
        if len(json_chat_data) >= 3:
            if len(json_chat_data[4][0]) >= 4:
                if json_chat_data[4][0][4]:
                    for img in json_chat_data[4][0][4]:
                        images.append(img[0][0][0])
        results = {
            "content": json_chat_data[4][0][1][0],
            "conversation_id": json_chat_data[1][0],
            "response_id": json_chat_data[1][1],
            "factualityQueries": json_chat_data[3],
            "textQuery": json_chat_data[2][0] if json_chat_data[2] is not None else "",
            "choices": [{"id": i[0], "content": i[1]} for i in json_chat_data[4]],
            "images": images,
        }
        self.conversation_id = results["conversation_id"]
        self.response_id = results["response_id"]
        self.choice_id = results["choices"][0]["id"]
        self._reqid += 100000
        return results


@register_engine
@singleton
class BardEngine(LLMChat):
    name = ("Bard", "bard")

    def __init__(
        self,
    ) -> None:
        super().__init__(device="api")

        session_id = os.environ.get("GOOGLE_BARD_SESSION_ID", None)
        if session_id is None:
            raise Exception(
                "GOOGLE_BARD_SESSION_ID environment variable not found. Please add it to your environment variables.",
            )
        self._chatbot = Chatbot(session_id=session_id)

    def call(self, conversation: Conversation, n_completions: int = 1, **kwargs: Any) -> List[ConversationTurn]:
        if len(conversation.turns) != 1:
            raise AssertionError("BARD conversations must have exactly one turn.")

        # Extract the last user turn from the conversation.
        user_turn = conversation.turns[-1]
        new_prompt = user_turn.text
        if user_turn.speaker in (Speaker.SYSTEM, Speaker.AI):
            raise AssertionError("The last turn in the conversation must be from the user.")

        outputs = []
        for _ in range(n_completions):
            logging.info("Sending prompt to BARD engine: %s", new_prompt)
            outputs.append(self._chatbot.ask(new_prompt)["content"])

        return [ConversationTurn(speaker=Speaker.AI, text=output) for output in outputs]
